@@ -35,6 +35,13 @@
 #include "gtkintl.h"
 #include "gtkbindings.h"
 
+typedef struct _ResponseData ResponseData;
+
+struct _ResponseData
+{
+  gint response_id;
+};
+
 static void gtk_dialog_class_init (GtkDialogClass *klass);
 static void gtk_dialog_init       (GtkDialog      *dialog);
 
@@ -60,6 +67,8 @@ static void gtk_dialog_map               (GtkWidget        *widget);
 
 static void gtk_dialog_close             (GtkDialog        *dialog);
 
+static ResponseData* get_response_data   (GtkWidget        *widget);
+
 enum {
   PROP_0,
   PROP_HAS_SEPARATOR
@@ -345,15 +354,42 @@ gtk_dialog_style_set (GtkWidget *widget,
   update_spacings (GTK_DIALOG (widget));
 }
 
+static gboolean
+dialog_has_cancel (GtkDialog *dialog)
+{
+  GList *children, *tmp_list;
+  gboolean ret = FALSE;
+      
+  children = gtk_container_get_children (GTK_CONTAINER (dialog->action_area));
+
+  for (tmp_list = children; tmp_list; tmp_list = tmp_list->next)
+    {
+      ResponseData *rd = get_response_data (tmp_list->data);
+      
+      if (rd && rd->response_id == GTK_RESPONSE_CANCEL)
+	{
+	  ret = TRUE;
+	  break;
+	}
+    }
+
+  g_list_free (children);
+
+  return ret;
+}
+
 static void
 gtk_dialog_close (GtkDialog *dialog)
 {
   /* Synthesize delete_event to close dialog. */
   
-  GdkEvent *event = gdk_event_new (GDK_DELETE);
-  GtkWidget *widget;
+  GtkWidget *widget = GTK_WIDGET (dialog);
+  GdkEvent *event;
 
-  widget = GTK_WIDGET (dialog);
+  if (!dialog_has_cancel (dialog))
+    return;
+
+  event = gdk_event_new (GDK_DELETE);
   
   event->any.window = g_object_ref (widget->window);
   event->any.send_event = TRUE;
@@ -459,13 +495,6 @@ gtk_dialog_new_with_buttons (const gchar    *title,
   return GTK_WIDGET (dialog);
 }
 
-typedef struct _ResponseData ResponseData;
-
-struct _ResponseData
-{
-  gint response_id;
-};
-
 static ResponseData*
 get_response_data (GtkWidget *widget)
 {
@@ -598,9 +627,9 @@ gtk_dialog_add_button (GtkDialog   *dialog,
 }
 
 static void
-gtk_dialog_add_buttons_valist(GtkDialog      *dialog,
-                              const gchar    *first_button_text,
-                              va_list         args)
+gtk_dialog_add_buttons_valist (GtkDialog      *dialog,
+                               const gchar    *first_button_text,
+                               va_list         args)
 {
   const gchar* text;
   gint response_id;
@@ -717,12 +746,7 @@ gtk_dialog_set_default_response (GtkDialog *dialog,
                                             "gtk-dialog-response-data");
 
       if (rd && rd->response_id == response_id)
-	{
-	  gtk_widget_grab_default (widget);
-	  
-	  if (!GTK_WINDOW (dialog)->focus_widget)
-	    gtk_widget_grab_focus (widget);
-	}
+	gtk_widget_grab_default (widget);
 	    
       tmp_list = g_list_next (tmp_list);
     }
@@ -920,13 +944,13 @@ gtk_dialog_run (GtkDialog *dialog)
 
   g_object_ref (dialog);
 
-  if (!GTK_WIDGET_VISIBLE (dialog))
-    gtk_widget_show (GTK_WIDGET (dialog));
-  
   was_modal = GTK_WINDOW (dialog)->modal;
   if (!was_modal)
     gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
 
+  if (!GTK_WIDGET_VISIBLE (dialog))
+    gtk_widget_show (GTK_WIDGET (dialog));
+  
   response_handler =
     g_signal_connect (dialog,
                       "response",
